@@ -76,10 +76,10 @@ class DBSessionInjector(object):
         self.engine = create_engine(dburl)
         from synoptic.datamodel import DataModel
         self.datamodel = DataModel()
-        self.datamodel.metadata.create_all(self.engine) 
+        self.datamodel.metadata.create_all(self.engine)
 
         from sqlalchemy.orm import sessionmaker
-        self.sessionmaker = sessionmaker(bind=self.engine, autoflush=True, 
+        self.sessionmaker = sessionmaker(bind=self.engine, autoflush=True,
                 autocommit=False)
 
         self.sub_app = sub_app
@@ -125,7 +125,7 @@ class ErrorMiddleware(object):
 
             start_response(status, response_headers, exc_info)
             return ["%s: %s" % (ex_type.__name__,str(ex_val))]
-        
+
 from paste.wsgiwrappers import WSGIRequest, WSGIResponse
 
 
@@ -159,16 +159,39 @@ class Request(WSGIRequest):
 
 
 
-        
+
 class ApplicationBase(object):
-    def __init__(self, table):
+    def __init__(self, table, allowed_networks=[]):
         import re
         self.table = [
                 (re.compile(match), dest)
                 for match, dest in table
                 ]
 
+        self.allowed_networks = allowed_networks
+
     def __call__(self, environ, start_response):
+        if self.allowed_networks:
+            from paste.httpexceptions import HTTPForbidden
+            try:
+                remote_addr_str = environ["REMOTE_ADDR"]
+            except KeyError:
+                raise HTTPForbidden("REMOTE_ADDR not found, "
+                        "but allowed_networks was specified")
+
+            from ipaddr import IPAddress
+            remote_addr = IPAddress(remote_addr_str)
+
+            allowed = False
+
+            for an in self.allowed_networks:
+                if remote_addr in an:
+                    allowed = True
+                    break
+
+            if not allowed:
+                raise HTTPForbidden("Requests from your address aren't allowed")
+
         for compiled_re, dest in self.table:
             match = compiled_re.search(environ["PATH_INFO"])
             if match is not None:
@@ -185,7 +208,7 @@ class ApplicationBase(object):
 
 
 class Application(ApplicationBase):
-    def __init__(self, urlprefix="/"):
+    def __init__(self, urlprefix="/", allowed_networks=[]):
         import re
         re_prefix = "^"+re.escape(urlprefix)
 
@@ -194,17 +217,21 @@ class Application(ApplicationBase):
                     (r'$', self.index),
                     (r'timestamp/get_range$', self.http_get_tsrange),
                     (r'item/get_by_id$', self.http_get_item_by_id),
+                    (r'item/get_version_by_id$', self.http_get_item_version_by_id),
                     (r'items/get$', self.http_get_items),
                     (r'items/print$', self.http_print_items),
                     (r'items/export$', self.http_export_items),
+                    (r'item/history/get$', self.http_get_item_history),
                     (r'item/store$', self.http_store_item),
                     (r'item/reorder$', self.http_reorder_item),
-                    (r'tags/get$', self.http_get_tags),
+                    (r'tags/get_filter$', self.http_get_tags),
+                    (r'tags/get_for_query$', self.http_get_tags_for_query),
                     (r'tags/rename$', self.http_rename_tag),
                     (r'app/get_all_js$', self.http_get_all_js),
                     (r'app/quit$', self.http_quit),
                     (r'static/([-_/a-zA-Z0-9.]+)$', self.serve_static),
-                    ]])
+                    ]],
+                allowed_networks=allowed_networks)
 
         WSGIRequest.defaults["charset"] = "utf-8"
         self.quit_func = None
@@ -215,9 +242,7 @@ class Application(ApplicationBase):
     # tools -------------------------------------------------------------------
     def item_to_json(self, item):
         result = item.as_json()
-
         result['contents_html'] = item.contents_html()
-        result['title'] = None
         return result
 
     def get_itemversions_for_request(self, request):
@@ -246,10 +271,10 @@ class Application(ApplicationBase):
         from sqlalchemy.sql import select
         iv_and_t_query = select([itemversions_query, model.tags],
                 from_obj=[itemversions_query
-                    .outerjoin(model.itemversions_tags, 
+                    .outerjoin(model.itemversions_tags,
                         itemversions_query.c.id
                         ==model.itemversions_tags.c.itemversion_id)
-                    .outerjoin(model.tags, 
+                    .outerjoin(model.tags,
                         model.itemversions_tags.c.tag_id
                         ==model.tags.c.id)
                     ],
@@ -262,7 +287,7 @@ class Application(ApplicationBase):
                 last_id = row[itemversions_query.c.id]
                 result.append(
                         {"id": row[itemversions_query.c.item_id],
-                            "title": None,
+                            "version_id": row[itemversions_query.c.id],
                             "contents": row[itemversions_query.c.contents],
                             "contents_html": ItemVersion.htmlize(
                                 row[itemversions_query.c.contents]),
@@ -297,8 +322,8 @@ class Application(ApplicationBase):
                     }),
                 mimetype="text/plain")
 
-    def get_tags_with_usecounts(self, session, model, parsed_query=None, 
-            max_timestamp=None):
+    def get_tags_with_usecounts(self, session, model, parsed_query=None,
+            max_timestamp=None, startswith=None, limit=None):
         from sqlalchemy.sql import select, and_, or_, not_, func
 
         if parsed_query is not None:
@@ -309,31 +334,52 @@ class Application(ApplicationBase):
                     from_obj=[get_current_itemversions_join(
                         model, max_timestamp)
                         ])
-                    
+
         itemversions_q = itemversions_q.alias("currentversions")
 
         # twuc_q stands for tags_with_usecount_query
         twuc_q = (
                 select([
-                    model.tags.c.name, 
+                    model.tags.c.name,
                     func.count(itemversions_q.c.id).label("use_count"),
                         ],
                     from_obj=[model.itemversions_tags
-                        .join(itemversions_q, 
+                        .join(itemversions_q,
                             itemversions_q.c.id==model.itemversions_tags.c.itemversion_id)
                         .join(model.tags)
                         ])
                 )
 
+        if startswith is None or not startswith.startswith("."):
+            twuc_q = twuc_q.where(~model.tags.c.name.startswith("."))
+
+        if startswith:
+            twuc_q = twuc_q.where(model.tags.c.name.startswith(startswith))
+
         twuc_q = (twuc_q
                 .group_by(model.tags.c.id)
                 .having(func.count(itemversions_q.c.id)>0)
                 .order_by(model.tags.c.name)
                 )
 
+        if limit is not None:
+            twuc_q = twuc_q.limit(limit)
+
         return session.execute(twuc_q)
 
     def http_get_tags(self, request):
+        query = request.GET.get("q", "")
+        limit = request.GET.get("limit", None)
+
+        q = self.get_tags_with_usecounts(
+                request.dbsession, request.datamodel,
+                startswith=query, limit=limit)
+
+        return request.respond(
+                "\n".join(row.name for row in q)+"\n",
+                mimetype="text/plain")
+
+    def http_get_tags_for_query(self, request):
         if "max_timestamp" in request.GET:
             max_timestamp = float(request.GET["max_timestamp"])
         else:
@@ -360,7 +406,7 @@ class Application(ApplicationBase):
         tags = [list(row) for row in result]
 
         return request.respond(
-                dumps({ 
+                dumps({
                 "tags": tags,
                 "query_tags": query_tags,
                 }),
@@ -375,7 +421,7 @@ class Application(ApplicationBase):
 
         tag = request.dbsession.query(Tag).filter_by(name=old_name).one()
         new_tag_query = request.dbsession.query(Tag).filter_by(name=new_name)
-        
+
         if new_tag_query.count():
             raise ValueError, "tag already exsits"
 
@@ -408,6 +454,20 @@ class Application(ApplicationBase):
         return request.respond(dumps(self.item_to_json(query.first())),
                 mimetype="text/plain")
 
+    def http_get_item_version_by_id(self, request):
+        query = (
+                request.dbsession.query(ItemVersion)
+                .filter_by(id=int(request.GET["version_id"]))
+                .limit(1))
+
+        if query.count() == 0:
+            from paste.httpexceptions import HTTPNotFound
+            raise HTTPNotFound()
+
+        from simplejson import dumps
+        return request.respond(dumps(self.item_to_json(query.first())),
+                mimetype="text/plain")
+
     def http_get_items(self, request):
         qry = request.GET.get("query", "")
 
@@ -420,7 +480,7 @@ class Application(ApplicationBase):
             max_timestamp = None
 
         json_items = self.get_json_items(
-                request.dbsession, request.datamodel, 
+                request.dbsession, request.datamodel,
                 parsed_query, max_timestamp)
 
         if qry != "":
@@ -443,7 +503,7 @@ class Application(ApplicationBase):
             else:
                 max_timestamp = None
 
-            tags = [list(row) 
+            tags = [list(row)
                     for row in self.get_tags_with_usecounts(
                         request.dbsession, request.datamodel,
                         max_timestamp=max_timestamp)]
@@ -467,7 +527,7 @@ class Application(ApplicationBase):
                 printpage({
                     "title": "Synoptic Printout",
                     "body": "<hr/>".join(
-                        '<div class="itemcontents">%s</div>' % v.contents_html() 
+                        '<div class="itemcontents">%s</div>' % v.contents_html()
                         for v in versions),
                     }))
 
@@ -483,6 +543,25 @@ class Application(ApplicationBase):
                     for v in versions),
                 mimetype="text/plain;charset=utf-8")
 
+    def http_get_item_history(self, request):
+        item_id = int(request.GET["item_id"])
+
+        json = [
+                {
+                    "version_id": iv.id, 
+                    "timestamp":iv.timestamp,
+                    "is_current":False,
+                    "contents":iv.contents,
+                    }
+                for iv in request.dbsession.query(ItemVersion)
+                .filter_by(item_id=item_id)
+                .order_by(ItemVersion.timestamp.desc())]
+        json[0]["is_current"] = True
+
+        from simplejson import dumps
+        return request.respond(dumps(json),
+                mimetype="text/plain")
+
     def http_store_item(self, request):
         from simplejson import loads, dumps
         data = loads(request.POST["json"])
@@ -506,7 +585,7 @@ class Application(ApplicationBase):
         else:
             voh = None
 
-        itemversion = store_itemversion(request.dbsession, 
+        itemversion = store_itemversion(request.dbsession,
                 data["contents"], data["tags"], data["id"])
 
         request.dbsession.commit() # fills in the item_id
@@ -545,6 +624,7 @@ class Application(ApplicationBase):
           "jquery.bgiframe.js",
           "jquery.dimensions.js",
           "jquery.contextmenu.r2.js",
+          "jquery.autocomplete.js",
           "jquery-ui/ui.core.js",
           "jquery-ui/ui.slider.js",
           "jquery-ui/ui.tabs.js",
@@ -558,14 +638,14 @@ class Application(ApplicationBase):
         sep = "/* %s */\n" % (75*"-")
         return request.respond("".join(
             "%s/* %s */\n%s%s" % (sep, fn, sep, get_static_file(fn)[0])
-            for fn in all_js_filenames), 
+            for fn in all_js_filenames),
             mimetype="text/javascript")
 
     def http_quit(self, request):
         if self.quit_func is not None:
             self.quit_func()
             return request.respond(
-                    "Thanks for using synoptic.", 
+                    "Thanks for using synoptic.",
                     mimetype="text/plain")
         else:
             raise HTTPForbidden()
